import copy
from contextlib import nullcontext as does_not_raise

import pytest

from jsonpath_ng.ext import parse


@pytest.mark.parametrize(
    "string, initial_data, expected_result",
    (
        ("$.foo", {}, {"foo": 42}),
        ("$.foo.bar", {}, {"foo": {"bar": 42}}),
        ("$.foo[0]", {}, {"foo": [42]}),
        ("$.foo[1]", {}, {"foo": [{}, 42]}),
        ("$.foo[0].bar", {}, {"foo": [{"bar": 42}]}),
        ("$.foo[1].bar", {}, {"foo": [{}, {"bar": 42}]}),
        ("$.foo[0][0]", {}, {"foo": [[42]]}),
        ("$.foo[1][1]", {}, {"foo": [{}, [{}, 42]]}),
        ("foo[0]", {}, {"foo": [42]}),
        ("foo[1]", {}, {"foo": [{}, 42]}),
        ("foo", {}, {"foo": 42}),
        #
        # Initial data can be a list if we expect a list back.
        ("[0]", [], [42]),
        ("[1]", [], [{}, 42]),
        #
        # Convert initial data to a list, if necessary.
        ("[0]", {}, [42]),
        ("[1]", {}, [{}, 42]),
        #
        (
            'foo[?bar="baz"].qux',
            {
                "foo": [
                    {"bar": "baz"},
                    {"bar": "bizzle"},
                ]
            },
            {"foo": [{"bar": "baz", "qux": 42}, {"bar": "bizzle"}]},
        ),
        ("[1].foo", [{"foo": 1}, {"bar": 2}], [{"foo": 1}, {"foo": 42, "bar": 2}]),
    ),
)
def test_update_or_create(string, initial_data, expected_result):
    jsonpath = parse(string)
    result = jsonpath.update_or_create(initial_data, 42)
    assert result == expected_result


@pytest.mark.parametrize(
    "string, initial_data, expectation",
    (
        # Slice not supported
        ("foo[0:1]", {}, does_not_raise()),
        #
        # Filter does not create items to meet criteria
        ('foo[?bar="baz"].qux', {}, does_not_raise()),
        #
        # Does not convert initial data to a dictionary
        ("foo", [], pytest.raises(TypeError)),
    ),
)
def test_unsupported_classes(string, initial_data, expectation):
    copied_initial_data = copy.copy(initial_data)
    jsonpath = parse(string)
    with expectation:
        result = jsonpath.update_or_create(initial_data, 42)
        assert result != copied_initial_data



def assert_value_equality(results, expected_values):
    """Assert equality between two objects.

    *results* must be a list of results as returned by `.find()` methods.

    If *expected_values* is a list, then value equality and ordering will be checked.
    If *expected_values* is a set, value equality and container length will be checked.
    Otherwise, the value of the results will be compared to the expected values.
    """

    left_values = [result.value for result in results]
    if isinstance(expected_values, list):
        assert left_values == expected_values
    elif isinstance(expected_values, set):
        assert len(left_values) == len(expected_values)
        assert set(left_values) == expected_values
    else:
        assert results[0].value == expected_values


def assert_full_path_equality(results, expected_full_paths):
    """Assert equality between two objects.

    *results* must be a list or set of results as returned by `.find()` methods.

    If *expected_full_paths* is a list, then path equality and ordering will be checked.
    If *expected_full_paths* is a set, then path equality and length will be checked.
    """

    full_paths = [str(result.full_path) for result in results]
    if isinstance(expected_full_paths, list):
        assert full_paths == expected_full_paths, full_paths
    else:  # isinstance(expected_full_paths, set):
        assert len(full_paths) == len(expected_full_paths)
        assert set(full_paths) == expected_full_paths



import pytest

from jsonpath_ng.ext import parse
from jsonpath_ng.ext.filter import Expression, Filter
from jsonpath_ng.jsonpath import Child, Descendants, Fields, Index, Root, Slice, This


@pytest.mark.parametrize(
    "string, parsed",
    [
        # The authors of all books in the store
        (
            "$.store.book[*].author",
            Child(
                Child(Child(Child(Root(), Fields("store")), Fields("book")), Slice()),
                Fields("author"),
            ),
        ),
        #
        # All authors
        ("$..author", Descendants(Root(), Fields("author"))),
        #
        # All things in the store
        ("$.store.*", Child(Child(Root(), Fields("store")), Fields("*"))),
        #
        # The price of everything in the store
        (
            "$.store..price",
            Descendants(Child(Root(), Fields("store")), Fields("price")),
        ),
        #
        # The third book
        ("$..book[2]", Child(Descendants(Root(), Fields("book")), Index(2))),
        #
        # The last book in order
        # "$..book[(@.length-1)]"  # Not implemented
        ("$..book[-1:]", Child(Descendants(Root(), Fields("book")), Slice(start=-1))),
        #
        # The first two books
        # "$..book[0,1]"  # Not implemented
        ("$..book[:2]", Child(Descendants(Root(), Fields("book")), Slice(end=2))),
        #
        # Filter all books with an ISBN
        (
            "$..book[?(@.isbn)]",
            Child(
                Descendants(Root(), Fields("book")),
                Filter([Expression(Child(This(), Fields("isbn")), None, None)]),
            ),
        ),
        #
        # Filter all books cheaper than 10
        (
            "$..book[?(@.price<10)]",
            Child(
                Descendants(Root(), Fields("book")),
                Filter([Expression(Child(This(), Fields("price")), "<", 10)]),
            ),
        ),
        #
        # All members of JSON structure
        ("$..*", Descendants(Root(), Fields("*"))),
    ],
)
def test_goessner_examples(string, parsed):
    """
    Test Stefan Goessner's `examples`_

    .. _examples: https://goessner.net/articles/JsonPath/index.html#e3
    """
    assert parse(string, debug=True) == parsed


def test_attribute_and_dict_syntax():
    """Verify that attribute and dict syntax result in identical parse trees."""

    assert parse("$.store.book[0].title") == parse("$['store']['book'][0]['title']")
